import inspect

from typing import Callable

import logging
from logging import handlers as logging_handlers

from .config import get_config_content
from .utils import ensure_dir
from .exceptions import InvalidOption, InvalidConfig, DuplicatedHandler


class LogProvider:
    """
    Get a LogmeLogger object for decorated classes and functions/methods

    *LogmeLogger object is provided as a object property of this LogProvider*
    """

    def __init__(self, decorated: Callable, config: str=None, name: str=None):
        """
        Initialization of Logger Provider, all the optional arguments should be passed from logme.log()

        Required: scope
        Optional: config, name.
                (logger configuration will resolve to the default when not specified,
                *given logme.ini file exists in root dir*)

        """

        self.decorated = decorated

        # Get the module object of the decorated object
        module_obj = inspect.getmodule(self.decorated)

        logger_name = name if name else module_obj.__name__

        config_dict = get_config_content(module_obj.__file__, name=config)

        self.logger = LogmeLogger(logger_name, config_dict)


class ModuleLogger:
    """
    Resolves to a LogmeLogger object based on the config and name passed from logme.log

    """
    def __init__(self, frame: int, config: str=None, name: str=None):
        """

        :param frame: frame number of the caller stack (1 if direct instantiate, 2 if instantiated from logme.log
        :param name: name of the logger
        :param config: configuration of the logger

        """
        module_frame = inspect.stack()[frame]
        module_obj = inspect.getmodule(module_frame.frame)

        logger_name = name if name else module_obj.__name__
        config_dict = get_config_content(module_frame.filename, name=config)

        self.logger = LogmeLogger(logger_name, config_dict)

    def __getattr__(self, attr):
        """
        Delegate the attributes and methods of self.logger to self
        """
        return getattr(self.logger, attr)


# ---------------------------------------------------------------------------
# Logger Object
# ---------------------------------------------------------------------------


class LogmeLogger:
    """
    Get a logger object with configured handlers

    """
    def __init__(self, name: str, config: dict):
        """
        :param name: name of the logger
        :param config: configuration of the logger
        """

        self.name = name
        self.config = config

        self._set_master_properties()

        self._set_handlers()

    def __getattr__(self, attr):
        """
        Delegate all the attributes and methods of logger to LogmeLogger Object
        """
        try:
            return getattr(self.logger, attr)
        except AttributeError:
            raise AttributeError(f"LogmeLogger object has no attribute '{attr}'.")

    @property
    def logger(self):
        logger = logging.getLogger(self.name)
        logger.setLevel(self.master_level)

        return logger

    def _set_master_properties(self):
        master_properties = {
            'master_formatter': logging.Formatter(self.config['formatter'], style='{'),
            'master_level': getattr(logging, self.config['level'].upper()),
            'handler_names':  [i for i in self.config.keys() if 'Handler' in i]

        }

        for k, v in master_properties.items():
            setattr(self, k, v)

    def _set_handlers(self):
        """
        Iterate through the config dict, set the active handlers
        """
        for handler_name in self.handler_names:

            if self.config[handler_name]['active'] is False:
                continue

            handler = self._get_handler(handler_name)

            level = self.config[handler_name].get('level')
            formatter = self.config[handler_name].get('formatter')
            self._config_handler(handler, formatter=formatter, level=level)

            if not self._handler_exist(handler):
                self.logger.addHandler(handler)

    def _get_handler(self, handler_name: str) -> logging.Handler:
        """ Get Handler object by name """

        try:
            handler_class = getattr(logging, handler_name)
        except AttributeError:
            handler_class = getattr(logging_handlers, handler_name)

        if handler_name == 'FileHandler':
            try:
                filename = self.config[handler_name]['filename']
                ensure_dir(filename)

                handler = handler_class(filename=filename)
            except TypeError:  # filename is None
                raise ValueError(f"file path for the {handler_name} must not be None")
        else:
            try:
                parse_args = {k: v for k, v in self.config[handler_name].items()
                              if k not in ['active', 'level', 'formatter']}

                handler = handler_class(**parse_args)
            except KeyError:
                raise InvalidConfig(f"{handler_name} is not configured in the logme.ini file, "
                                    f"current available handlers: {self.handler_names}")

        return handler

    def _config_handler(self, handler, level: str=None, formatter: str=None):
        """
        Configure the handler's level and formatter
        """
        # Set level
        if level:
            handler.setLevel(getattr(logging, level.upper()))
        else:
            handler.setLevel(self.master_level)

        # Set formatter
        if formatter:
            formatter_object = logging.Formatter(formatter, style='{')
            handler.setFormatter(formatter_object)
        else:
            handler.setFormatter(self.master_formatter)

    def _handler_exist(self, handler: logging.Handler) -> bool:
        """
        Check if logging handler already exists
        """
        for i in self.logger.handlers:
            handler_class = i.__class__
            handler_attr = self._get_handler_attr(handler)
            exist_attr = self._get_handler_attr(i)

            # Using type(handler) here because FileHandler inherit from StreamHandler,
            # and isinstance() will be True is checking a FileHandler object with StreamHandler class
            if type(handler) is handler_class and handler_attr == exist_attr:
                return True

        return False

    def add_handler(self, handler_type: str, formatter: str=None,
                    level: str=None, allow_duplicate: bool=False, **kwargs):
        """
        Add the handler to self.logger on adhoc basis

        :param handler_type: The type of handler, e.g. StreamHandler, SocketHandler
        :param formatter: formatter to be passed to the handler
        :param level: Level for the handler
        :param allow_duplicate: *USE WITH CAUTION*, this allows duplication of handlers in the same logger

        :param kwargs: arguments to be passed to the handler class

        """
        handler_class = getattr(logging, handler_type)

        handler = handler_class(**kwargs)
        self._config_handler(handler, level=level, formatter=formatter)

        if not self._handler_exist(handler) or allow_duplicate:
            self.logger.addHandler(handler)
        else:
            raise DuplicatedHandler(f"{handler_class} with the exact same configuration already exists, "
                                    f"add allow_duplicate=True to allow.")

    def _get_handler_attr(self, handler: logging.Handler) -> dict:
        """
        Get the *string* attribute of the handler
        """
        handler_attr = {'formatter': handler.formatter._fmt}

        for k, v in handler.__dict__.items():
            if isinstance(v, str) or str(v).isdigit():
                handler_attr[k] = v

        return handler_attr

    def reset_config(self, config_name: str=None, config: dict=None, name: str=None):
        """
        Used for alternative config. This is normally used for module scope loggers

        :param config_name: config names specified in the config
        :param config: config dict
        :param name: logger name to be changed to
        """
        # Check if the arguments are valid
        if config and config_name:
            raise InvalidOption("Can only set keyword argument of either 'config_name' or 'config', not both.")
        if not config and not config_name:
            raise InvalidOption("must specify one of 'config_name' or 'config'.")

        caller_file_path = inspect.getframeinfo(inspect.currentframe().f_back).filename
        if config_name:
            self.config = get_config_content(caller_file_path, config_name)
        else:
            self.config = config

        if name:
            self.name = name

        # Remove existing handlers and configure new
        for i in self.logger.handlers:
            self.logger.removeHandler(i)

        self._set_master_properties()
        self._set_handlers()



import click

from pathlib import Path
from collections import defaultdict
from configparser import ConfigParser
from contextlib import contextmanager

from .config import read_config
from .exceptions import LogmeError
from .utils import dict_to_conf, ensure_dir

_command_options = {
    'project_root': click.option('--project-root', '-p',
                                 help='The project root, where the logme.ini is, or where you want to place it. '
                                      'Relative/absolute path',
                                 default='.'),
    'level': click.option('--level', '-lvl',
                          help='the logging level',
                          default='DEBUG'),
    'formatter': click.option('--formatter', '-f',
                              help='the logging formatter',
                              default='{asctime} - {name} - {levelname} - {message}'),
    'log_path': click.option('--log-path', '-lp',
                             help='the filename where you want to store your log',
                             default=None),
}


def add_options(options: list=None):

    """
    *decorator*

    Combining _command_options as one single decorator,

    - options can be passed to specify which option to use
    - if option is not specified, it will combine all the option

    """

    def add_options_wrapper(command_func):
        if not options:
            for option in _command_options.values():
                command_func = option(command_func)
        else:
            for option in options:
                command_func = _command_options[option](command_func)

        return command_func

    return add_options_wrapper


@click.group(context_settings={'help_option_names': ['-h', '--help']})
def cli():
    """
    *entry point*

    Simple init:

        >>> logme init

    To make a new log config file in specified dir:

        >>> logme init -p /var/workspace/my_project -mk

    Make a new logger config in existing config file:

        >>> logme logger my_logger

    """


@cli.command()
@click.option('--mkdir', '-mk',
              help='Create the directory in which logme.ini resides.',
              is_flag=True)
@add_options()
@click.pass_context
def init(ctx, project_root, mkdir, level, formatter, log_path):
    """
    Entry point.

    Command to set up a logme config file with master logger configuration

    """
    conf_content = get_tpl('logme', level=level, formatter=formatter, filename=log_path)

    config = get_config(conf_content)

    abs_path = Path(project_root).resolve()
    conf_location = abs_path.joinpath('logme.ini')

    if not abs_path.exists():
        if not mkdir:
            raise NotADirectoryError(f"{abs_path.parent.resolve() / project_root} does not exist. If you'd "
                                     f"like to make the directory, please use '-mk' flag.")
        else:
            abs_path.mkdir(parents=True, exist_ok=True)

    with conf_location.open('w') as conf:
        config.write(conf)


@cli.command()
@click.argument('name', required=1)
@add_options()
@click.pass_context
def add(ctx, project_root, name, level, formatter, log_path):
    """
    Command for adding a logger configuration in a logme.ini file. Assuming logme.ini exists
    """
    with ensure_conf_exist(project_root) as logme_conf:

        validate_conf(name, logme_conf)

        conf_content = get_tpl(name, level=level, formatter=formatter, filename=log_path)
        config = get_config(conf_content)

        # check if section already exist
        with logme_conf.open('a') as conf:
            config.write(conf)


@cli.command()
@click.argument('name', required=1)
@add_options(['project_root'])
@click.pass_context
def remove(ctx, name, project_root):
    """
    Command for removing a logger configuration in a logme.ini file.
    logme configuration cannot be removed
    """

    if name == 'logme':
        raise LogmeError("'logme' master logger configuration cannot be removed!")

    with ensure_conf_exist(project_root) as logme_conf:

        config = read_config(logme_conf)
        config.remove_section(name)

        with logme_conf.open('w+') as conf:
            config.write(conf)


@contextmanager
def ensure_conf_exist(project_root: str) -> Path:
    """
    Ensure logme.ini file exists, if it does, return the file path as pathlib.Path object,

    :param project_root: project_root where logme.ini is.
    :raises: LogmeError, if the configuration file does not exist
    """

    logme_conf = Path(project_root) / 'logme.ini'

    if logme_conf.exists():
        yield logme_conf
    else:
        raise FileNotFoundError(f"log me config file does not exist in {Path.cwd()}, "
                                f"if you'd like to initialize logme in this directory, please type 'logme init'.")


def validate_conf(name, ini_file_path):
    """
    Helper function for 'logme logger' command,
    ensure the logger name passed in does not already exist in the ini file

    :param name:
    :param ini_file_path:
    :return:
    """
    sections = read_config(ini_file_path).sections()

    if name in sections:
        raise LogmeError(f"'{name}' logging config already exists in config file: {ini_file_path}")
    elif 'logme' not in sections:
        raise LogmeError(f"{ini_file_path} is not a valid logme.ini file")


def get_config(conf_content: dict) -> ConfigParser:
    """

    :param conf_content: nested dict

    :return: configpaser.Configparser
    """
    config = ConfigParser()
    # preserve casing
    config.optionxform = str

    for k, v in conf_content.items():
        config[k] = v

    return config


def get_tpl(name: str, **kwargs: str) -> dict:

    """
    Get the template dict for the logger configuration

    :param name: str, name of the logger
    :param kwargs: keys = ['level', 'formatter', 'filename']

    :return: dict
    """

    check_options(**kwargs)

    config = defaultdict(dict)

    logger_template = {
            'level': 'DEBUG',
            'formatter': None,
            'StreamHandler': {
                'active': True,
                'level': 'DEBUG',
            },
            'FileHandler': {
                'active': False,
                'level': 'DEBUG',
                'filename': 'mylogpath/foo.log',
            },
            'NullHandler': {
                'active': False,
                'level': 'NOTSET'
            },
        }

    map_template(logger_template, kwargs)

    # flatten config dictionary
    config[name] = dict(dict_to_conf(logger_template))

    return dict(config)


def map_template(template: dict, input_: dict) -> None:
    """
    Update the template dict with the input, if input value exists

    *This updates the original dict passed in*

    :param template: dict
    :param input_: keys = ['level', 'formatter', 'filename']
    """
    for k, v in template.items():
        config_val = input_.get(k)

        if isinstance(v, dict) and k != 'NullHandler':
            map_template(v, input_)

        if config_val:
            template[k] = config_val.upper() if k == 'level' else config_val


def check_options(**kwargs):

    """
    check the level and log_path options passed, if the log_path doesn't exist, make the dir

    :param kwargs: keys = ['level', 'log_path']

    :raises: LogmeError, if invalid level passed
    """

    allowed_levels = ['critical', 'error', 'warning', 'info', 'debug', 'notset',
                      '50', '40', '30', '20', '10', '0']

    if kwargs.get('level') and kwargs['level'].lower() not in allowed_levels:
        raise LogmeError(f"{kwargs['level']} is not allowed, "
                         f"please specify the following: {allowed_levels}")

    # Make the directory and the log file
    if kwargs.get('filename'):
        ensure_dir(kwargs['filename'])




"""




"""

import inspect
from typing import Callable

from functools import wraps

from .cli import cli

from .utils import check_scope
from .exceptions import LogmeError, MisMatchScope
from .providers import LogProvider, ModuleLogger


def log(scope: str=None, config: str=None, name: str=None):
    """
    Returns a decorator or logger object

    :param scope: scope of the logger
    :param config: name of the logging config specified in logme.ini
    :param name: name of the logger

    """
    if isinstance(scope, str) and scope.lower() == 'module':
        return ModuleLogger(frame=2, config=config, name=name)

    if callable(scope):
        return _get_logger_decorator(scope, config=config, name=name)
    else:
        if scope:
            check_scope(scope.lower(), ['class', 'function', 'module'])

        def wrapper(decorated):
            return _get_logger_decorator(decorated, config=config, name=name, scope=scope)
        return wrapper


def _get_logger_decorator(callable_: callable, config: str=None, name: str=None, scope: str=None) -> Callable:
    """
    Get the logger decorator based on what kind of callable is being passed, class | function
            - Inject a keyword arg to function/method
            - Inject an attribute 'logger' to a class based decorator

    """
    if inspect.isclass(callable_):
        if scope and scope != 'class':
            raise MisMatchScope(f"{callable_} is a class, cannot be assigned to a '{scope}' scope")

        provider = LogProvider(callable_, config=config, name=name)
        callable_.logger = provider.logger
        return callable_

    if inspect.isfunction(callable_):
        if scope and scope != 'function':
            raise MisMatchScope(f"{callable_} is a class, cannot be assigned to a '{scope}' scope")

        provider = LogProvider(callable_, config=config, name=name)

        @wraps(callable_)
        def wrapper(*args, **kwargs):
            return callable_(*args, **kwargs, logger=provider.logger)

        return wrapper

    raise LogmeError("'{callable_}' must be a 'class' or a 'function'.")


__version__ = '1.0.5'



from importlib.machinery import SourceFileLoader
from pathlib import Path

from setuptools import setup, find_packages

module_path = Path(__file__).parent / Path('logme/__init__.py')
logme_module = SourceFileLoader('logme', str(module_path)).load_module()

version = logme_module.__version__

requires = [
    'click',
]

with open('README.rst', 'r', encoding='utf-8') as f:
    readme = f.read()

setup(
    name='logme',
    packages=find_packages(exclude=['tests*']),
    install_requires=requires,
    version=version,
    description='package for easy logging',
    long_description=readme,
    author='Luna Chen',
    url='https://github.com/BNMetrics/logme',
    author_email='luna@bnmetrics.com',
    keywords=['logging', 'cli'],
    python_requires='>=3',
    entry_points={'console_scripts': ['logme=logme:cli']},
    license='Apache 2.0',
)
